#ifndef ICLIENT_HH
#define ICLIENT_HH

#include <iostream>
#include <sstream>

#include <errno.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/tcp.h>
#include <netdb.h> 
#include <unistd.h>
#include <poll.h>

#include <imsg.hh>
#include <message.hh>
#include <msgConstants.hh>

class iclient {

private:
	int sock;

public:
	enum received{STAT_ERR, STAT_EMPT, STAT_RECV};

	iclient() {}
	
	~iclient(){
		disconnect();
	}

	bool connect(const std::string& serverName, int portNumber){
		if(isConnected()){
			//If already coonected, drop the last coonection
			disconnect();
		}
		
		//Set up the socket
		struct addrinfo * address, *current;
		struct addrinfo hints;
		memset(&hints, 0, sizeof(struct addrinfo)); //Create empty hint
		hints.ai_family=AF_UNSPEC; //For both IPv4 and IPv6
		hints.ai_socktype=SOCK_STREAM;
		std::stringstream port;
		port<<portNumber;
		int result=getaddrinfo(serverName.c_str(), port.str().c_str(), &hints, &address);
		if(result!=0){
			if(result==EAI_SYSTEM){
				int errnum=errno;
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<strerror(errnum)<<std::endl;
				return false;
			}else{
				std::cerr<<"Error while creating the address :"<<std::endl;
				std::cerr<<gai_strerror(result)<<std::endl;
				return false;
			}
		}
		
		for (current=address; current != NULL; current=current->ai_next) {
			sock=socket(current->ai_family, current->ai_socktype, current->ai_protocol);
			if(sock==-1){
				continue;
			}
			
			//Allow multiple connexions
			int reuse_addr=1;
			if(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof(reuse_addr))!=0){
				close(sock);
				continue;
			}
			
			if(::connect(sock, current->ai_addr, current->ai_addrlen) == 0)
				break;
			
			close(sock);
		}
		
		freeaddrinfo(address);
		
		if(current==NULL || sock==-1 || result==-1){
			int errnum=errno;
			std::cerr<<"Error while creating socket :"<<std::endl;
			std::cerr<<strerror(errnum)<<std::endl;
			return false;
		}

		return true;
	}

	void subscribe(const std::string& name) const{
		message post;
		post.encode(DEFAULT_SERVER_NAME, name.c_str());
		send(post);
	}

	bool isConnected() const{
		//Check the socket state
		int error;
		socklen_t len=sizeof(error);
		int retval=getsockopt(sock, SOL_SOCKET, SO_ERROR, &error, &len);
		return(retval==0);
		//If retval == 0 the socket is up, otherwise it is disconnected (it allow to detect only some errors)
	}

	void disconnect(){
		close(sock);		
	}

	bool checkReadSocket(bool blocking) const{
		if(!isConnected())
			return false;
		int timeout=0; //Non blocking
		if(blocking){
			timeout=-1; //Wait forever
		}
		
		struct pollfd fd;
		memset(&fd, 0 , sizeof(fd));
		fd.fd=sock;
		fd.events=POLLIN;
		
		int result=poll(&fd, 1, timeout);
		
		if(result==0){ //Timed out
			return false;
		}
		
		if((fd.revents & POLLERR) || (fd.revents & POLLHUP) || (fd.revents & POLLNVAL)){
			std::cerr<<"Error on the socket"<<std::endl;
		}
		if(result<0){
			std::cerr<<"Error on the socket while waiting for a message"<<std::endl;
			std::cerr<<strerror(errno)<<std::endl;
		}
		return(fd.revents & POLLIN);
	}
	
	bool checkReadSocketTimeout(int ms) const{
		if(!isConnected())
			return false;
		struct pollfd fd;
		memset(&fd, 0 , sizeof(fd));
		fd.fd=sock;
		fd.events=POLLIN;
		
		int result=poll(&fd, 1, ms);
		
		if((fd.revents & POLLERR) || (fd.revents & POLLHUP) || (fd.revents & POLLNVAL)){
			std::cerr<<"Error on the socket"<<std::endl;
		}
		if(result<0){
			std::cerr<<"Error on the socket while waiting for a message"<<std::endl;
			std::cerr<<strerror(errno)<<std::endl;
		}
		return(fd.revents & POLLIN);
	}

	bool send(const message& data) const{
		if(!isConnected())
			return false;
		imsg msg;
		if(!msg.encode(data.data()))
			return false;
		ssize_t n;
		size_t tot = 0;
		size_t bleft = msg.length();
		while(tot < msg.length()) {
			n = ::send(sock, msg.data() + tot, bleft, MSG_NOSIGNAL);
			if(n < 1){
				std::cerr<<"Error while sending the message"<<std::endl;
				std::cerr<<"\t"<<strerror(errno)<<std::endl;
				break;
			}
			tot += n;
			bleft -= n;
		}
		return (tot == msg.length());
	}

	std::pair<received, message> receive() const{
		message result;
		imsg msg;
		ssize_t res;
		char header[msg.hsize()];
		errno=0;

		if(!isConnected()){
			std::cerr<<"Peer disconnected"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}

		res = ::recv(sock, header, msg.hsize(), 0);
		if(res<1){
			std::cerr<<"Error on the socket while receiving a message :"<<std::endl;
			std::cerr<<strerror(errno)<<std::endl;
			return std::make_pair(STAT_ERR, result);
		} else if(res==0){
			std::cerr<<"Peer disconnected"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}else
		if(res != static_cast<ssize_t>(msg.hsize())){
			std::cerr<<"Error while reading socket, header does not have the correct length"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeHeader(header);
		char body[msg.blength()];
		res = ::recv(sock, body, msg.blength(), MSG_WAITALL);
		if(res != static_cast<ssize_t>(msg.blength())){
			std::cerr<<"Error while reading socket, body does not have the correct length"<<std::endl;
			return std::make_pair(STAT_ERR, result);
		}
		msg.decodeBody(body);
		result.decode(msg.body());
		return std::make_pair(STAT_RECV, result);
	}
};

#endif
